/* -*- mode:c; tab-width:4; c-basic-offset:4; -*-
 *
 * This file is part of maemo-security-certman
 *
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * Contact: Juhani Mäkelä <ext-juhani.3.makela@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include <stdio.h>
#include "/usr/include/expat_external.h"
#include <nss.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <linux/limits.h>
#include <sys/fcntl.h>
#include <cert.h>
#include <secmod.h>
#include <pk11pub.h>
#include <p12.h>
#include <certdb.h>
#include <openssl/pem.h>
#include <aegis_common.h>
#include <aegis_certman.h>
#include <x509_container.h>

#undef AEGIS_ERROR
#define AEGIS_ERROR(format,args...) \
	do {\
        fprintf(stderr, "%s(%d): ERROR: " format "\n", bare_file_name(__FILE__), __LINE__ ,##args);\
	} while (0)

static int
cert_info(X509* cert)
{
	aegis::x509_container cc(cert);
    printf("\n%s%s\n\tSubject: %s\n\tIssuer: %s\n\tNot before: %s\n\tNot after: %s\n", 
           cc.issuer_name_nid(NID_organizationName)->c_str(), 
           cc.is_valid_timewise()?"":" (EXPIRED)",
           cc.subject_name(), cc.is_self_signed()?"(Self signed)":cc.issuer_name(),
           cc.not_before_timestamp()->c_str(), cc.not_after_timestamp()->c_str());
    return(cc.is_valid_timewise());
}

/*
 * PK11_ListCertsInSlot uses this?
 */
enum pw_type {
    PW_NONE = 0,
    PW_FROMFILE = 1,
    PW_PLAINTEXT = 2,
    PW_EXTERNAL = 3
};

typedef struct {
	enum pw_type source;
	char* data;
} secuPWData;

#define DEFAULT_CFG_DIR  "/home/user/.netscape"
#define DEFAULT_DLL_NAME "/usr/lib/libaegis_certman.so.0.0.0"
#define DEFAULT_MOD_NAME "maemo-security-certman"
#define NSS_FLAGS        "trustOrder=70 \
 slotParams=0x6a5=[slotFlags=PublicCerts,rootFlags]"

static char cfg_dir  [PATH_MAX] = DEFAULT_CFG_DIR;
static char mod_name [PATH_MAX] = DEFAULT_MOD_NAME;
static char dll_name [PATH_MAX] = DEFAULT_DLL_NAME;
static char tgt_dir  [PATH_MAX] = "";

static enum op_mode_t {
	mode_update, 
	mode_info, 
	mode_remove, 
	mode_remove_internal,
	mode_export,
	mode_maemo_export,
    mode_none
} op_mode = mode_none;


static void
usage(void)
{
	printf("%s\n", "Usage:\n"
           "nsscfg -c config-dir [-m module] [-l library]\n"
           " to install a new securitymodule or\n"
           "nsscfg -e <tgtdir> or nsscfg -E\n"
           " to export all certificates from the default certificate store"
           );
}

static int
rm_file(int ordnr, void* name, void* context)
{
    char fullname[256];
    snprintf(fullname, sizeof(fullname), "%s/%s", (const char*) context, (const char*)name);
    fprintf(stderr, "Remove '%s'\n", (const char*)name);
    unlink(fullname);
    return 0;
}


int
main(int argc, char* argv[])
{
	SECStatus rv;
	SECMODModule* mod = NULL;
	signed char a;
	int rc;

	AEGIS_DEBUG(1, "started");

	while (1) {
		a = getopt(argc, argv, "c:m:l:ir:Re::E");
		if (0 > a)
			break;
		switch (a) 
			{
			case 'c':
                op_mode = mode_update;
				strncpy(cfg_dir, optarg, sizeof(cfg_dir));
				break;
			case 'm':
                op_mode = mode_update;
				strncpy(mod_name, optarg, sizeof(mod_name));
				break;
			case 'l':
                op_mode = mode_update;
				strncpy(dll_name, optarg, sizeof(dll_name));
				break;
			case 'i':
				op_mode = mode_info;
				break;
			case 'r':
				op_mode = mode_remove;
				strncpy(mod_name, optarg, sizeof(mod_name));
				break;
			case 'R':
				op_mode = mode_remove_internal;
				break;
			case 'e':
				op_mode = mode_export;
                if (optarg) 
                    strncpy(tgt_dir, optarg, sizeof(tgt_dir));
				break;
			case 'E':
                if (0 != getuid()) {
                    AEGIS_ERROR("must be run as root in this mode");
                    return -1;
                }
				op_mode = mode_maemo_export;
                strncpy(tgt_dir, "/etc/certs/common-ca", sizeof(tgt_dir));
				break;
			default:
				AEGIS_DEBUG(1, "Invalid option '%hd'", a);
				usage();
				return -1;
			}
	}

    switch (op_mode) {
        case mode_update:
        case mode_remove:
            /*
             * Update secmodule deb, initialize in default mode
             * with a specific root directory.
             */
            AEGIS_DEBUG(1, "Configure %s", cfg_dir);
            if (!directory_exists(cfg_dir)) {
                rc = create_directory(cfg_dir, 0755);
                if (0 != rc) {
                    AEGIS_ERROR("Could not create dir '%s' (%d)", 
                                   cfg_dir, rc);
                    return -1;
                } else
                    AEGIS_DEBUG(1, "Created dir '%s'", cfg_dir);
            }

            rv = NSS_Initialize(cfg_dir, "", "", "", 0);
            AEGIS_DEBUG(1, "NSS_Initialize returned %d", rv);
            if (SECSuccess != rv) {
                AEGIS_ERROR("Failed to initialize NSS library for '%s' (%d)", 
                               cfg_dir, rv);
                return -1;
            }
            break;

        case mode_info:
            /*
             * Query preinstalled certificate set. Initialize in 
             * read-only mode.
             */
            rv = NSS_Initialize(cfg_dir, "", "", "", NSS_INIT_READONLY);
            AEGIS_DEBUG(1, "NSS_Initialize returned %d", rv);
            if (SECSuccess != rv) {
                AEGIS_ERROR("Failed to initialize NSS library for '%s' (%d)", 
                               cfg_dir, rv);
                return -1;
            }
            break;

        case mode_export:
        case mode_maemo_export:
            /*
             * Query preinstalled certificate set. Initialize in 
             * read-only mode.
             */
            rv = NSS_Initialize(cfg_dir, "", "", "", NSS_INIT_READONLY | NSS_INIT_NOMODDB);
            AEGIS_DEBUG(1, "NSS_Initialize returned %d", rv);
            if (SECSuccess != rv) {
                AEGIS_ERROR("Failed to initialize NSS library for '%s' (%d)", 
                               cfg_dir, rv);
                return -1;
            }
            break;

        default:
            usage();
            return 0;
    }

	if (mode_update == op_mode) {
		mod = SECMOD_FindModule(mod_name);
		if (mod) {
			AEGIS_DEBUG(1, "Updating '%s'", mod_name);
			mod->dllName = dll_name;
			rv = SECMOD_UpdateModule(mod);
			AEGIS_DEBUG(1, "SECMOD_UpdateModule returned %d", rv);
		} else {
			AEGIS_DEBUG(1, "Adding '%s'", mod_name);
			rv = SECMOD_AddNewModuleEx(mod_name, dll_name, 0, 0, NULL, (char*)NSS_FLAGS);
			AEGIS_DEBUG(1, "SECMOD_AddNewModuleEx returned %d", rv);
		}
		if (SECSuccess == rv)
			printf("Added %s in %s\n", mod_name, cfg_dir);
		else
			printf("ERROR: %d in adding %s to %s\n", rv, mod_name, cfg_dir);

	} else if (mode_info == op_mode) {
		SECMODModuleList* m_list = SECMOD_GetDefaultModuleList();
		while (m_list) {
			if (m_list->module) {
				printf("Module: %s\nLibrary: %s\nisCritical=%s\nisModuleDB=%s\n"
					   "moduleDBOnly=%s\ntrustOrder=%d\n",
					   m_list->module->commonName,
					   m_list->module->dllName,
					   m_list->module->isCritical?"yes":"no",
					   m_list->module->isModuleDB?"yes":"no",
					   m_list->module->moduleDBOnly?"yes":"no",
					   m_list->module->trustOrder
					   );
			}
			if (m_list->next) {
				printf("\n");
				m_list = m_list->next;
			} else
				break;
		}

	} else if (mode_remove == op_mode) {
		int type = 0;
		rv = SECMOD_DeleteModule(mod_name, &type);
		if (SECSuccess == rv)
			printf("Removed '%s'\n", mod_name);
		else
			AEGIS_ERROR("Cannot remove '%s' (%d)\n", mod_name, rv);

	} else if (mode_remove_internal == op_mode) {
		if (SECMOD_CanDeleteInternalModule()) {
			mod = SECMOD_GetInternalModule();
			if (mod && mod->commonName) {
				rv = SECMOD_DeleteInternalModule(mod->commonName);
				if (SECSuccess == rv)
					printf("Removed internal module '%s'\n", mod->commonName);
				else
					AEGIS_ERROR("Cannot remove internal module '%s' (%d)", 
							mod->commonName, rv);
			} else
				AEGIS_ERROR("No internal module exists");
		} else
			AEGIS_ERROR("Cannot delete internal module (no permission)");

	} else if (mode_export == op_mode || mode_maemo_export == op_mode) {

        const char domain_name [] = "common-ca";

		CERTCertDBHandle *dbh = CERT_GetDefaultCertDB();
		CERTCertList *certs;
		CERTCertListNode *node;
		secuPWData  pwdata = { PW_NONE, 0 };
		int rc, ccount = 0;
        domain_handle common_ca_domain = AEGIS_CERTMAN_DOMAIN_NONE;

		if (0 < strlen(tgt_dir) && !directory_exists(tgt_dir)) {
            printf("Export the Mozilla certificate set to %s...\n", tgt_dir);
			create_directory(tgt_dir, 0755);
        } else if (mode_maemo_export == op_mode) {
            printf("Create new common-ca store from the Mozilla certificate set\n");
            iterate_files(tgt_dir, ".*", rm_file, tgt_dir);
            aegis_certman_open(NULL);
            rc = aegis_certman_open_domain(domain_name, AEGIS_CERTMAN_DOMAIN_SHARED, 
                                              &common_ca_domain);
            if (0 != rc) {
                AEGIS_ERROR("Cannot open domain '%s' (%d)", domain_name, rc);
                return -1;
            }
        }

		if (!dbh) {
			AEGIS_ERROR("CERT_GetDefaultCertDB returned NULL");
			goto shutdown;
		}

		certs = PK11_ListCerts(PK11CertListAll, &pwdata);
		if (certs) {
            AEGIS_DEBUG(1, "Got certs list");

			for (node = CERT_LIST_HEAD(certs); 
				 !CERT_LIST_END(node,certs);
				 node = CERT_LIST_NEXT(node)) 
			{
                X509 *cert;
                char cert_name[256];

				ccount++;
                AEGIS_DEBUG(1, "Handling cert %d", ccount);

                cert = d2i_X509(NULL, (const unsigned char**)&node->cert->derCert.data, 
                                node->cert->derCert.len);

                AEGIS_DEBUG(1, "Converted to OpenSSL format", ccount);

                if (NULL != cert) {
                    rc = aegis_certman_get_nickname(cert, cert_name, sizeof(cert_name));
                    if (0 != rc) {
                        AEGIS_ERROR("Invalid certificate (%d/%d)", ccount, rc);
                        continue;
                    }
                } else {
                    AEGIS_ERROR("Invalid certificate (%d)", ccount);
                    continue;
                }

                if (mode_export == op_mode) {
                    aegis_key_id cert_key;
                    char cert_key_str[AEGIS_KEY_ID_STR_LEN];

                    if (cert_info(cert)) {
                        aegis_certman_get_key_id(cert, cert_key);
                        aegis_certman_key_id_to_str(cert_key, cert_key_str, sizeof(cert_key_str));

                        if (0 < strlen(tgt_dir)) {
                            char file_name[PATH_MAX];
                            FILE* fh;

                            snprintf(file_name, sizeof(file_name), "%s/%s.pem", tgt_dir, cert_key_str);
                            fh = fopen(file_name, "wb");
                            if (PEM_write_X509(fh, cert)) {
                                printf("%s\n\t=> %s\n", cert_name, file_name);
                                fclose(fh);
                            }
                        }
                    }

                } else {
                    if (cert_info(cert)) {
                        rc = aegis_certman_add_cert(common_ca_domain, cert);
                        if (0 != rc) {
                            AEGIS_ERROR("Cannot add '%s' to '%s' (%d)", cert_name, domain_name, rc);
                        }
                    }
                }
                X509_free(cert);
			}
			CERT_DestroyCertList(certs);
		}
		printf("There are %d certs in the database\n", ccount);

        if (AEGIS_CERTMAN_DOMAIN_NONE != common_ca_domain) {
            rc = aegis_certman_close_domain(common_ca_domain);
            if (0 != rc) {
                AEGIS_ERROR("closing domain '%s' failed (%d)", domain_name, rc);
            }
        }
	}
		
 shutdown:
	rv = NSS_Shutdown();
	AEGIS_DEBUG(1, "NSS_Shutdown returned %d", rv);
    aegis_certman_close(NULL);

	return 0;
}
