/* -*- mode:c++; tab-width:4; c-basic-offset:4; -*-
 *
 * This file is part of aegis-certman
 *
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * Contact: Juhani Mäkelä <ext-juhani.3.makela@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include <aegis_certman.h>
#include <aegis_crypto.h>
#include <dbus/dbus.h>
#include <errno.h>
#include <openssl/evp.h>
#include <pthread.h>
#include <string.h>
#include <sys/time.h>

// Used connection type (session / system bus)
#define DBUS_CONNECTION DBUS_BUS_SYSTEM

int dbus_exchange(DBusMessage *, DBusMessage **, long);

typedef void (*NOTIFY_KEY)(void*, int, EVP_PKEY *);
typedef void (*NOTIFY_CERT)(void*, int);

struct _AsyncUserData
{
    const char* credentials;
    void* state;
    void* notify;
};

struct _Timeout
{
    DBusTimeout * timeout;
    unsigned long start;
    unsigned long remains;
    dbus_bool_t enabled;
    struct _Timeout * next;
};

struct _Connection
{
    int setup;
    pthread_mutex_t mutex;
    DBusConnection * connection;
    struct _Timeout * timeouts;
};

static struct _Connection _GUIConnection = {0, PTHREAD_MUTEX_INITIALIZER};

static unsigned long
_timeout_now(void)
{
    struct timeval t;
    unsigned long timestamp;

    gettimeofday(&t, NULL);

    // We ignore the overflow since it is not significant for our
    // intervals (we only compare timestamps for a difference in seconds, not years)
    timestamp = (unsigned long)t.tv_sec * 1000;
    timestamp += (unsigned long)t.tv_usec / 1000;

    return timestamp;
}

static dbus_bool_t
_timeout_add(DBusTimeout * timeout, void * data)
{
    struct _Timeout * new_timeout = dbus_malloc(sizeof(struct _Timeout));
    struct _Timeout * end = _GUIConnection.timeouts;

    AEGIS_DEBUG(1, "%s(%d)", __func__, dbus_timeout_get_interval(timeout));

    if ( !new_timeout )
        return FALSE;

    // Set up the timeout
    new_timeout->timeout = timeout;
    new_timeout->start = _timeout_now();
    new_timeout->remains = dbus_timeout_get_interval(timeout);
    new_timeout->enabled = TRUE; // It should be enabled or there is no need to add it
    new_timeout->next = NULL;

    // Append the new timeout
    if ( !end )
        _GUIConnection.timeouts = new_timeout;
    else {
        // Find the end of the list
        while (end->next)
            end = end->next;

        end->next = new_timeout;
    }

    return TRUE;
}

static void
_timeout_remove(DBusTimeout * timeout, void * data)
{
    struct _Timeout * prev = _GUIConnection.timeouts;
    struct _Timeout * removed;

    AEGIS_DEBUG(1, "%s(%d)", __func__, dbus_timeout_get_interval(timeout));

    if ( !prev )
        return;

    if (prev->timeout == timeout) {
        removed = prev;
        _GUIConnection.timeouts = prev->next;
    }
    else {
        while (prev->next)
            if (prev->next->timeout == timeout)
                break;
            else
                prev = prev->next;

        if ( !prev->next ) // Not our timeout
            return;

        removed = prev->next;
        prev->next = removed->next;
    }

    dbus_free(removed);
}

static void
_timeout_toggled(DBusTimeout * timeout, void * data)
{
    struct _Timeout * t = _GUIConnection.timeouts;

    AEGIS_DEBUG(1, "%s(%d): %d", __func__, dbus_timeout_get_interval(timeout), dbus_timeout_get_enabled(timeout));

    while (t)
        if (t->timeout != timeout)
            t = t->next;
        else {
            t->enabled = dbus_timeout_get_enabled(timeout);

            if (t->enabled) {
                t->start = _timeout_now();
                t->remains = (unsigned long)dbus_timeout_get_interval(timeout);
            }
            break;
        }
}

static int
_gui_connection_setup(void)
{
    DBusError error;

    if (_GUIConnection.setup)
        return 0;

    dbus_threads_init_default();
    dbus_error_init(&error);
    _GUIConnection.connection = dbus_bus_get(DBUS_CONNECTION, &error);
    if (NULL == _GUIConnection.connection) {
        AEGIS_ERROR("Failed to open bus (%s:%s)",
                    error.name, error.message);
        dbus_error_free(&error);
        return EAGAIN;
    }

    if ( !dbus_connection_set_timeout_functions(_GUIConnection.connection,
                                                &_timeout_add,
                                                &_timeout_remove,
                                                &_timeout_toggled,
                                                NULL,
                                                NULL) )
    {
        dbus_connection_unref(_GUIConnection.connection);
        AEGIS_ERROR("Failed to set up connection: no memory");
        return ENOMEM;
    }

    _GUIConnection.setup = 1;

    return 0;
}

static void
_timeout_process(void)
{
    struct _Timeout * timeout = _GUIConnection.timeouts;

    while (timeout) {
        struct _Timeout * next = timeout->next;

        if (timeout->enabled) {
            unsigned long now = _timeout_now();
            unsigned long elapsed;

            elapsed = (now < timeout->start) ? timeout->remains : now - timeout->start;
            // The first is possible if system clock has been reset into the past

            timeout->start = now;
            if (elapsed > timeout->remains)
                timeout->remains = 0;
            else
                timeout->remains -= elapsed;

            if ( !timeout->remains ) {
                timeout->remains = dbus_timeout_get_interval(timeout->timeout); // Restart
                AEGIS_DEBUG(1, "%s: timeout elapsed", __func__);
                dbus_timeout_handle(timeout->timeout);
            }
        }

        // next is used to prevent a situation when the triggered timeout has
        // been removed inside a dbus_timeout_handle()
        timeout = next;
    }
}

static int
_send_request(DBusMessage * message, DBusPendingCall ** pcall, long timeout)
{
    int result = 0;
    DBusConnection * connection = _GUIConnection.connection;
    dbus_bool_t send_start = 0;

    *pcall = 0;

    send_start = dbus_connection_send_with_reply(connection,
                                                 message,
                                                 pcall,
                                                 timeout * 1000);
    if (send_start && *pcall)
        dbus_connection_flush(connection);
    else {
        AEGIS_ERROR("Failed to send message (%s)",
                    (send_start) ? "com.nokia.certman not present" : "out of memory");
        result = ENOMSG;
    }

    return result;
}



static int
_decrypt_reply_private_key(DBusMessage * reply, const char * with_credentials, EVP_PKEY ** the_key)
{
    DBusError error;
    int rc = 0;
    char *resp = NULL;

    dbus_error_init(&error);
    if (!dbus_message_get_args(reply, &error, DBUS_TYPE_STRING, &resp, DBUS_TYPE_INVALID)) {
        AEGIS_ERROR("Failed to parse reply (%s:%s)", error.name, error.message);
        dbus_error_free(&error);
        return EBADMSG;
    }

    AEGIS_DEBUG(1, "%s: got '%s'", __func__, resp);

    RAWDATA_PTR rawdata_encrypted = NULL;
    RAWDATA_PTR plaintext = NULL;
    size_t len, reslen;
    len = base64_decode(resp, &rawdata_encrypted);
    if (0 < len) {
        aegis_crypto_result res = aegis_crypto_decrypt(rawdata_encrypted,
                                                       len,
                                                       with_credentials,
                                                       &plaintext,
                                                       &reslen);
        if (aegis_crypto_ok == res) {
            unsigned char *tbuf = (unsigned char*)plaintext;
            AEGIS_DEBUG(1, "%s: got response of len %d bytes", __func__, reslen);
            *the_key = d2i_AutoPrivateKey(NULL, (const unsigned char**)&tbuf, reslen);
            if (*the_key) {
                AEGIS_DEBUG(1, "%s: encoded to private key", __func__);
                rc = 0;
            } else {
                AEGIS_DEBUG(1, "%s: d2i_AutoPrivateKey failed, returned NULL", __func__);
                rc = EACCES;
            }
        } else {
            AEGIS_DEBUG(1, "%s: could not decrypt result", __func__);
            rc = EACCES;
        }
        if (rawdata_encrypted)
            aegis_crypto_free(rawdata_encrypted);
        if (plaintext)
            aegis_crypto_free(plaintext);
    } else {
        AEGIS_DEBUG(1, "%s: could not decode result", __func__);
        rc = ENOMEM;
    }

    return rc;
}

static int
_compose_request_private_key(aegis_key_id with_id,
                     const char* with_credentials,
                     DBusMessage ** message)
{
    char key_id [AEGIS_KEY_ID_STR_LEN], *key_id_a, *creds_a;

    *message = NULL;
    if (0 != aegis_certman_key_id_to_str(with_id, key_id, sizeof(key_id))) {
        return EINVAL;
    }

    *message = dbus_message_new_method_call("com.nokia.certman",
                                           "/request",
                                           "com.nokia.certman",
                                           "GetPrivateKey");

    if (NULL == *message) {
        AEGIS_ERROR("Failed to create message (%s)", strerror(errno));
        return ENOMEM;
    }

    key_id_a = key_id;
    creds_a = (char*)with_credentials;

    errno = 0;
    if (!dbus_message_append_args( *message,
                                  DBUS_TYPE_STRING,
                                  &key_id_a,
                                  DBUS_TYPE_STRING,
                                  &creds_a,
                                  DBUS_TYPE_INVALID))
    {
        AEGIS_ERROR("Failed to append arguments (%s)", strerror(errno));
        return errno; // Is this really set by dbus_message_append_args()?
    }

    return 0;
}

static int
_compose_request_certificate(X509 *cert, DBusMessage ** message)
{
    unsigned char *cert_as_der = NULL;
    char *cert_as_pem = NULL;
    size_t len;
    int rc = 0;


    *message = dbus_message_new_method_call("com.nokia.certman",
                                           "/request",
                                           "com.nokia.certman",
                                           "CheckCertificate");

    len = i2d_X509(cert, &cert_as_der);
    if (0 == len || NULL == cert_as_der) {
        AEGIS_ERROR("Failed to convert cert to DER (%d)", len);
        return EPERM;
    }

    AEGIS_DEBUG(1, "%s: cert %p len %d", __func__, cert_as_der, len);
    cert_as_pem = base64_encode(cert_as_der, len);
    OPENSSL_free(cert_as_der);

    if (NULL == cert_as_pem) {
        AEGIS_ERROR("Failed to convert cert to PEM (%d)", len);
        return EPERM;
    }

    rc = dbus_message_append_args( *message, DBUS_TYPE_STRING, &cert_as_pem, DBUS_TYPE_INVALID );
    aegis_crypto_free(cert_as_pem);

    if (!rc) {
        AEGIS_ERROR("Failed to append PEM encoded certificate");
        return ENOMEM;
    }

    return 0;
}

static void _notify_private_key_received(DBusPendingCall * pcall, void * udata)
{
    struct _AsyncUserData * user_data = (struct _AsyncUserData *)udata;
    DBusMessage * reply;
    EVP_PKEY * key = NULL;
    NOTIFY_KEY notify = (NOTIFY_KEY)user_data->notify;
    int rc = 0;

    AEGIS_ENTER;

    // It is assumed that this function is always called with a valid udata allocated in
    // aegis_certman_gui_get_private_key_async(). The memory is freed automatically when
    // pcall is unreferenced.

    // We always get a reply even if a timeout occurs or the connection is closed
    reply = dbus_pending_call_steal_reply(pcall);


    if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
        rc = ENOMSG;
    else
        rc = _decrypt_reply_private_key(reply, user_data->credentials, &key);

    if(notify)
        notify(user_data->state, rc, key);

    dbus_message_unref(reply);
    dbus_pending_call_unref(pcall); // user_data is freed at this point by dbus_free()
}

static void _notify_certificate_checked(DBusPendingCall * pcall, void * udata)
{
    struct _AsyncUserData * user_data = (struct _AsyncUserData *)udata;
    NOTIFY_CERT notify = (NOTIFY_CERT)user_data->notify;
    DBusMessage * reply;
    int rc = 0;

    AEGIS_ENTER;

    // It is assumed that this function is always called with a valid udata allocated in
    // aegis_certman_gui_check_certificate_async(). The memory is freed automatically when
    // pcall is unreferenced.

    // We always get a reply even if a timeout occurs or the connection is closed
    reply = dbus_pending_call_steal_reply(pcall);

    if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
        rc = ENOMSG;
    else {
        int approved = 0;
        DBusError error;

        dbus_error_init( &error );
        if( !dbus_message_get_args(reply, &error, DBUS_TYPE_BOOLEAN, &approved, DBUS_TYPE_INVALID) )
            rc = EBADMSG;
        else
            if( !approved )
                rc = EACCES;
    }

    if(notify)
        notify(user_data->state, rc);

    dbus_message_unref(reply);
    dbus_pending_call_unref(pcall); // user_data is freed at this point by dbus_free()
}

int
dbus_exchange(DBusMessage *message, DBusMessage **reply, long timeout)
{
    DBusConnection *connection = NULL;
    int result = 0;
    DBusError error;

    if (NULL == reply)
        return EINVAL;

    dbus_error_init(&error);
    connection = dbus_bus_get(DBUS_CONNECTION, &error);
    if (NULL == connection) {
        AEGIS_ERROR("Failed to open session bus (%s:%s)", 
                    error.name, error.message);
        dbus_error_free(&error);
        return EAGAIN;
    }
    
    *reply = dbus_connection_send_with_reply_and_block(connection,
                                                      message,
                                                      timeout * 1000,
                                                      &error);
    if (dbus_error_is_set(&error)) {
        AEGIS_ERROR("Failed to send message (%s:%s)", 
                    error.name, error.message);
        dbus_error_free(&error);
        result = ENOMSG;
    }
    if (connection)
        dbus_connection_unref(connection);
        
    return result;
}


int
aegis_certman_gui_check_certificate(X509 *cert, long timeout)
{
    DBusConnection *connection = NULL;
    DBusMessage *message = NULL, *reply = NULL;
    DBusError error;
    int approved = 0;
    int rc = 0;

    AEGIS_ENTER;

    if (NULL == cert) {
        AEGIS_ERROR("Invalid argument (cert == NULL)");
        return EINVAL;
    }

    dbus_threads_init_default(); // libdbus allows multiple inits
    dbus_error_init(&error);

    connection = dbus_bus_get(DBUS_CONNECTION, &error);
    if (NULL == connection) {
        AEGIS_ERROR("Failed to open system bus (%s:%s)",
                    error.name, error.message);
        rc = EAGAIN;
        goto finish;
    }

    rc = _compose_request_certificate(cert, &message);
    if (rc)
        goto finish;

    dbus_error_init(&error);
    reply = dbus_connection_send_with_reply_and_block(connection,
                                                      message,
                                                      timeout * 1000,
                                                      &error);
    if (dbus_error_is_set(&error)) {
        AEGIS_ERROR("Failed to send message (%s:%s)", 
                    error.name, error.message);
        rc = EAGAIN;
        goto finish;
    }

    dbus_error_init(&error);
    if (!dbus_message_get_args(reply, &error, DBUS_TYPE_BOOLEAN, &approved, DBUS_TYPE_INVALID)) {
        AEGIS_ERROR("Failed to parse reply (%s:%s)", error.name, error.message);
        rc = EACCES;
        goto finish;
    }

    if (!rc && !approved)
        rc = EACCES;

 finish:
    dbus_error_free(&error);
    if (reply)
        dbus_message_unref(reply);
    if (message)
        dbus_message_unref(message);
    if (connection)
        dbus_connection_unref(connection);

    return rc;
}

int
aegis_certman_gui_check_certificate_async(X509 *cert,
                                  long timeout,
                                  void (*notify)(void*, int),
                                  void* state) {
    DBusMessage *request = NULL;
    DBusPendingCall * pcall;
    int rc = 0;
    pthread_mutex_t * mutex = &_GUIConnection.mutex;

    AEGIS_ENTER;

    if(pthread_mutex_trylock(mutex))
        return EAGAIN;

    if ( !_GUIConnection.setup ) {
        rc = _gui_connection_setup();

        if (rc) {
            AEGIS_ERROR("Could not set up a connection");
            pthread_mutex_unlock(mutex);
            return rc;
        }
    }

    rc = _compose_request_certificate(cert, &request);
    if ( !rc )
        rc = _send_request(request, &pcall, timeout);
    else
        AEGIS_ERROR("Could not compose request");

    if ( !rc ) {
        struct _AsyncUserData * user_data = dbus_malloc(sizeof(struct _AsyncUserData));

        if (NULL == user_data) {
            dbus_pending_call_cancel(pcall); // The reply will just be ignored if it is received
            AEGIS_ERROR("Could not allocate memory for a pending call data");
            rc = ENOMEM;
        }
        else {
            user_data->credentials = NULL;
            user_data->notify = notify;
            user_data->state = state;

            if (dbus_pending_call_set_notify(pcall, &_notify_certificate_checked, user_data, &dbus_free) == FALSE) {
                dbus_pending_call_cancel(pcall); // The reply will just be ignored if it is received
                dbus_free(user_data);
                AEGIS_ERROR("Could not allocate memory in dbus_pending_call_set_notify()");
                rc = ENOMEM;
            }
        }
    }
    else
        AEGIS_ERROR("Could not send a request");

    if (request)
        dbus_message_unref(request);

    pthread_mutex_unlock(mutex);

    return rc;
}

int
aegis_certman_gui_get_private_key_async(aegis_key_id with_id,
                                  const char* with_credentials,
                                  long timeout,
                                  void (*notify)(void*, int, EVP_PKEY *),
                                  void* state) {
    DBusMessage *request = NULL;
    DBusPendingCall * pcall;
    int rc = 0;
    size_t creds_len;
    pthread_mutex_t * mutex = &_GUIConnection.mutex;

    AEGIS_ENTER;

    // A sanity check for credentials string
    if (NULL == with_credentials
            || (creds_len = strlen(with_credentials)) > 1024)
        return EINVAL;

    if(pthread_mutex_trylock(mutex))
        return EAGAIN;

    if ( !_GUIConnection.setup ) {
        rc = _gui_connection_setup();

        if (rc) {
            AEGIS_ERROR("Could not set up a connection");
            pthread_mutex_unlock(mutex);
            return rc;
        }
    }

    rc = _compose_request_private_key(with_id, with_credentials, &request);
    if ( !rc )
        rc = _send_request(request, &pcall, timeout);
    else
        AEGIS_ERROR("Could not compose request");

    if ( !rc ) {
        struct _AsyncUserData * user_data = dbus_malloc(
                sizeof(struct _AsyncUserData) + creds_len + 1);

        if (NULL == user_data) {
            dbus_pending_call_cancel(pcall); // The reply will just be ignored if it is received
            AEGIS_ERROR("Could not allocate memory for a pending call data");
            rc = ENOMEM;
        }
        else {
            char * creds = (char *)user_data + sizeof(struct _AsyncUserData);
            strcpy(creds, with_credentials);

            user_data->credentials = (const char *)creds;
            user_data->notify = notify;
            user_data->state = state;

            if (dbus_pending_call_set_notify(pcall,
                    &_notify_private_key_received, user_data, &dbus_free) == FALSE) {
                dbus_pending_call_cancel(pcall); // The reply will just
                                                 // be ignored if it is
                                                 // received
                dbus_free(user_data);
                AEGIS_ERROR("Could not allocate memory in dbus_pending_call_set_notify()");
                rc = ENOMEM;
            }
        }
    }
    else
        AEGIS_ERROR("Could not send a request");

    if (request)
        dbus_message_unref(request);

    pthread_mutex_unlock(mutex);

    return rc;
}

void
aegis_certman_gui_poll(long wait)
{
    pthread_mutex_t * mutex = &_GUIConnection.mutex;

    if( !_GUIConnection.setup )
        return;

    if (0 > wait)
        wait = 1;

    if(pthread_mutex_trylock(mutex))
        return;

    _timeout_process();
    dbus_connection_read_write_dispatch(_GUIConnection.connection, wait);

    pthread_mutex_unlock(mutex);
}

int
aegis_certman_gui_get_private_key(aegis_key_id with_id,
                                  const char* with_credentials,
                                  long timeout, 
                                  EVP_PKEY** the_key)
{
    DBusMessage *message = NULL, *reply = NULL;
    int rc = 0;

    AEGIS_ENTER;

    if (NULL == the_key || NULL == with_credentials)
    {
        return EINVAL;
    }
            
    dbus_threads_init_default(); // libdbus allows multiple inits

    rc = _compose_request_private_key(with_id, with_credentials, &message);
    if (rc)
        goto finish;

    rc = dbus_exchange(message, &reply, timeout);
    if (0 != rc) {
        goto finish;
    }

    rc = _decrypt_reply_private_key(reply, with_credentials, the_key);

 finish:
    if (reply)
        dbus_message_unref(reply);
    if (message)
        dbus_message_unref(message);
    return rc;
}
