/* -*- mode:c++; tab-width:4; c-basic-offset:4; -*-
 *
 * This file is part of aegis-certman
 *
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * Contact: Juhani Mäkelä <ext-juhani.3.makela@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

#include <aegis_certman.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <regex.h>

// STL headers
#include <string>
#include <vector>
// #include <dequeue>
#include <stack>
#include <map>
using namespace std;

// OpenSSL headers
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/pem.h>
#include <openssl/rand.h>
#include <openssl/pkcs12.h>
// #ifdef USE_BB5
// #include <libbb5.h>
// #endif
#include <aegis_common.h>
#include <aegis_crypto.h>
#include <aegis_storage.h>
#include "x509_container.h"

using namespace aegis;


/*
 * Storage name prefix and directory names
 */

static const char cert_storage_prefix [] = "certman";
static const char common_cert_dir     [] = "/var/lib/aegis/certs";
static const char priv_cert_dir       [] = "user";
static const char priv_keys_dir       [] = "keys";

/*
 * Directory access bits. Create all directories as
 * world writable so that all users can access and 
 * modify them. The protection is taken care by 
 * other means.
 */

#define PUBLIC_DIR_MODE  0777
#define PRIVATE_DIR_MODE 0777


// TODO: should this really be a public
// EVP_PKEY *root_pkey = NULL;
static int is_initialized = 0;

static char* cur_creds = NULL;


namespace aegis
{
    /**
	 * \brief A secure certificate container
	 */
	struct local_domain
	{
		storage* index;    ///< The secure storage containing the files
		string   dirname;  ///< The directory in which the actual files are
		int      domflags; ///< Type of the domain
	};
}


static bool
verify_cert(X509_STORE* ctx, X509* cert, bool allow_unknown_issuer)
{
	X509_STORE_CTX *csc;
	bool verify_ok;
	int rc;

	csc = X509_STORE_CTX_new();
	if (csc == NULL) {
		AEGIS_ERROR("cannot create new context");
		return(false);
	}

	rc = X509_STORE_CTX_init(csc, ctx, cert, NULL);
	if (rc == 0) {
		AEGIS_ERROR("cannot initialize new context");
		return(false);
	}

	verify_ok = (X509_verify_cert(csc) > 0);
	if (!verify_ok)
		AEGIS_DEBUG(1, "%s: verification failed: %s", __func__,
					X509_verify_cert_error_string(csc->error));

	if (allow_unknown_issuer && 
		X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY == csc->error)
		/*
		 * TODO: Make sure that this does not make openssl to ignore
		 * other, more severe errors.
		 */
		verify_ok = 1;

	X509_STORE_CTX_free(csc);

	return verify_ok;
}


static bool
load_certs(vector<string> &certnames,
		   bool do_verify,
		   X509_STORE* certs
) {
	map<string, x509_container*> cert_map;
	stack<x509_container*> temp;
	int error, i = 0;

	// TODO: Is this logic necessary at all now that the 
	// certificates have been divided to domains? After all,
	// the tool should not force verification if that's not
	// what the user wants.

	// Load self signed certificates directly to X509 store
	// and put the rest into a map for quick access

	for (size_t i = 0; i < certnames.size(); i++) {

		x509_container* cert = new x509_container(certnames[i].c_str());

		if (0 < strlen(cert->key_id())) {
	        if (cert->is_self_signed()) 
			{
				AEGIS_DEBUG(1, "self signed: %s", cert->subject_name());
				cert->m_verified = true;
				X509_STORE_add_cert(certs, cert->cert());
				delete(cert);

			} else {
				cert_map[cert->key_id()] = cert;
				AEGIS_DEBUG(1, "%s\n\tkey    %s\n\tissuer %s", 
							   cert->subject_name(),
							   cert->key_id(), 
							   cert->issuer_key_id());
			}

		} else {
			AEGIS_ERROR("Invalid certificate '%s'", cert->subject_name());
			delete(cert);
		}
	}

	/*
	 * Load and verify the rest of the certificates in the proper order
	 */
	for (map<string, x509_container*>::const_iterator ii = cert_map.begin();
		ii != cert_map.end();
		ii++) 
	{
		x509_container* cert = ii->second;
		x509_container* issuer;

		if (!cert) {
			AEGIS_DEBUG(0, "%s: what hell? (%s)", __func__, ii->first.c_str());
			continue;
		}
		AEGIS_DEBUG(2, "%s: iterate next (%d,%p,%d)", __func__, ++i, cert, 
					cert->m_verified);

		/* Find possible issuer for those stupid certificates
		 * that didn't specify it by a x509v3 extension. If none
		 * is found, never mind as then the issuer must already
		 * be in the certs-store.
		 */
		if (0 == strlen(cert->issuer_key_id())) {
			bool found_issuer = false;
			AEGIS_DEBUG(1, "Searching issuer for '%s'", cert->subject_name());
			for (map<string, x509_container*>::const_iterator jj = cert_map.begin();
				 jj != cert_map.end();
				 jj++) 
			{
				if (0 == strcmp(cert->issuer_name(), jj->second->subject_name())) {
					if (cert->is_issued_by(jj->second->cert(), &error)) {
						AEGIS_DEBUG(1, "Found issuer");
						found_issuer = true;
						cert->set_issuer(jj->second);
						break;
					} else {
						AEGIS_DEBUG(1, "Issuer names match but verification fails?");
					}
				} else {
					AEGIS_DEBUG(5, "'%s' != '%s'",  cert->issuer_name(), jj->second->subject_name());
				}
			}
			if (!found_issuer)
				AEGIS_DEBUG(1, "Issuer '%s' not found", cert->issuer_name());
		}

                /*
                 * Be really sure the certificate is not self-signed.
                 * If for some reason OpenSSL cannot decode the public
                 * key but the certificate is really self signed we end
                 * up in an endless loop
                 */
                if(!cert->m_verified && (0 == strcmp(cert->key_id(), cert->issuer_key_id()) ||
                                         0 == strcmp(cert->subject_name(), cert->issuer_name()))){
                        AEGIS_DEBUG(1, "self signed certificate %s couldn't be verified", cert->key_id());
                        continue;
                }

		if (!cert->m_verified) {
			temp.push(cert);

			/* 
			 * Verify issuers first, if any exist
			 */
			while (cert_map.count(cert->issuer_key_id())) {
				issuer = cert_map[cert->issuer_key_id()];
				if (issuer) {
					if (!issuer->m_verified) {
						AEGIS_DEBUG(1, "push %s", issuer->issuer_key_id());
						temp.push(issuer);
					} else {
						AEGIS_DEBUG(1, "issuer already verified");
						break;
					}
				} else {
					AEGIS_ERROR("cannot find issuer %s for %s", 
						  cert->issuer_key_id(),
						  cert->subject_name()); 
					return(false);
				}
                                /*
                                 * Make sure the chain is not longer than it's maximum size
                                 */
                                if (temp.size() > cert_map.size() + 1) {
                                        AEGIS_DEBUG(1, "Cross Signed certificate cycle detected");
                                        break;
                                }
				cert = issuer;
			}

			while (temp.size()) {
				cert = temp.top();
				temp.pop();
				if (verify_cert(certs, cert->cert(), !do_verify)) {
					AEGIS_DEBUG(2, "verified: %s", cert->subject_name());
					X509_STORE_add_cert(certs, cert->cert());
					cert->m_verified = true;
				} else {
					AEGIS_ERROR("%s verification fails", cert->subject_name());
				}
			}
		} else
			AEGIS_DEBUG(0, "Already verified");
	}
	AEGIS_DEBUG(2, "erasing map");
	for (
		map<string, x509_container*>::const_iterator ii = cert_map.begin();
		ii != cert_map.end();
		ii++
	) {
		x509_container* cert = ii->second;
		delete(cert);
	}
	cert_map.clear();
	return(true);
}


static void
local_storage_dir(string& to_this, const char* subarea)
{
    to_this.assign(common_cert_dir);
    to_this.append(PATH_SEP);
    to_this.append(subarea);
    to_this.append(PATH_SEP);
}


static void
decide_storage_name(const char* domain_name, int flags, string& dirname, string& storename)
{
	if (AEGIS_CERTMAN_DOMAIN_PRIVATE == flags) {
		// Make private name
		local_storage_dir(dirname, priv_cert_dir);
		storename.insert(0, ".");
		storename.insert(0, cert_storage_prefix);
		if (domain_name) {
			dirname.append(domain_name);
			storename.append(domain_name);
		}
		AEGIS_DEBUG(1, "\ndirname  = %s\nstorename = %s", dirname.c_str(), storename.c_str());
	} else {
		storename.assign(domain_name);
		storename.insert(0, ".");
		storename.insert(0, cert_storage_prefix);
		dirname.assign(common_cert_dir);
		dirname.append(PATH_SEP);
		dirname.append(domain_name);
		AEGIS_DEBUG(1, "\ndirname  = %s\nstorename = %s", dirname.c_str(), storename.c_str());
	}
}


static void
hex_to_key_id(const char* hstring, unsigned char* to_id)
{
	unsigned int val;

	memset(to_id, '\0', AEGIS_KEY_ID_LEN);
	for (int i = 0; i < 2*AEGIS_KEY_ID_LEN; i += 2) {
		if (0 == sscanf(hstring + i, "%02x", &val)) {
			AEGIS_ERROR("invalid key file name '%s' at %d", hstring, i);
		}
		*to_id++ = (unsigned char) val;
	}
}


static X509*
load_cert_from_file(const char* from_file)
{
	FILE* fp;
	X509* cert;

	fp = fopen(from_file, "r");
	if (!fp) {
		fprintf(stderr, "Cannot read '%s' (%d)\n", from_file, errno);
		return(0);
	}
	cert = PEM_read_X509(fp, NULL, 0, NULL);
	if (!cert) {
		fprintf(stderr, "Cannot read certificate from '%s'\n", from_file);
	}
	fclose(fp);
	return(cert);
}


static bool
make_unique_filename(X509* of_cert, struct local_domain* in_domain, string& to_string)
{
	bool result = true;
	aegis_key_id key_id;
	char nickname[255];

	AEGIS_ENTER;
	to_string.assign(in_domain->dirname.c_str());
	to_string.append(PATH_SEP);
	if (NULL == of_cert || 
		0 != aegis_certman_get_nickname(of_cert, nickname, sizeof(nickname))) 
	{
		AEGIS_ERROR("%s: invalid parameter", __func__);
		return false;
	}
	AEGIS_DEBUG(1, "%s: for '%s'", __func__, nickname);

	if (0 != aegis_certman_get_key_id(of_cert, key_id)) {
		AEGIS_ERROR("Cannot get key id out of certificate");
		return false;
	}

	for (int i = 0; result && i < 1000 ; i++) {
		string test_name;
		test_name.assign(to_string);
		append_hex(test_name, key_id, AEGIS_KEY_ID_LEN);
		if (0 < i) {
			char onbr[10];
			sprintf(onbr, "-%d", i);
			test_name.append(onbr);
		}
		test_name.append(".pem");
		AEGIS_DEBUG(1, "%s: testing '%s'", __func__, test_name.c_str());
		if (in_domain->index->contains_file(test_name.c_str())) {
			X509* ocert = load_cert_from_file(test_name.c_str());
			if (NULL != ocert) {
				if (0 == X509_cmp(of_cert, ocert)) {
					AEGIS_DEBUG(1, "%s: domain '%s' already contains equal certificate",
								__func__, in_domain->index->name());
					to_string.assign(test_name);
					result = false;
				} else {
					AEGIS_DEBUG(1, "%s: domain '%s' contains same key but different cert",
								__func__, in_domain->index->name());
				}
				X509_free(ocert);
			} else {
				to_string.assign(test_name);
				break;
			}
		} else {
			to_string.assign(test_name);
			break;
		}
	}
	AEGIS_DEBUG(1, "%s: %s (%s)", __func__, to_string.c_str(), result?"ok":"duplicate");
	return result;
}


#define MAX_TRIES 100

void
make_hash_filename(X509* of_cert, storage* pstore, const char* to_certfile, string &result)
{
	X509* lcert = of_cert;
	char hash_file_name[32];
	string full_name;
	long hash;
	int i;

	AEGIS_DEBUG(1, "%s: make hash to '%s'", __func__, to_certfile);
	if (NULL == lcert) {
		lcert = load_cert_from_file(to_certfile);
		if (NULL == lcert) {
			AEGIS_ERROR("%s: can't load '%s'",
						   __func__, to_certfile);
			return;
		}
	}
	hash = X509_subject_name_hash(lcert);
	if (of_cert != lcert)
		X509_free(lcert);
	
	for (i = 0; i < MAX_TRIES; i++) {
		snprintf(hash_file_name, sizeof(hash_file_name), "%08lx.%d", hash, i);
		if (!pstore->contains_link(hash_file_name))
			break;
	}
	if (MAX_TRIES == i) {
		AEGIS_ERROR("%s: %d colliding hash files for '%s'?",
					   __func__, i, to_certfile);
	} else {
		result.assign(hash_file_name);
	}
}


static int
store_key_to_file(aegis_key_id key_id, EVP_PKEY* key, char* passwd)
{
	string storage_file_name;
	FILE* outfile;
	int rc;
	
	if (0 == EVP_PKEY_size(key)) {
		AEGIS_ERROR("Invalid key size");
		return EINVAL;
	}

	local_storage_dir(storage_file_name, priv_keys_dir);
	create_directory(storage_file_name.c_str(), PRIVATE_DIR_MODE);
	append_hex(storage_file_name, key_id, AEGIS_KEY_ID_LEN);
	storage_file_name.append(".pem");

	outfile = fopen(storage_file_name.c_str(), "w");
	if (outfile) {
        /* Allow reading for other users too, to enable installing
         * keys in the image.
         */
		chmod(storage_file_name.c_str(), 0644);
		rc = PEM_write_PKCS8PrivateKey(outfile, key, 
                                       EVP_aes_128_cbc(), 
									   passwd, strlen(passwd), NULL, NULL);
		AEGIS_DEBUG(1, "Stored key to '%s', rc = %d", 
					   storage_file_name.c_str(), rc);
		fclose(outfile);
		return(0);
	} else {
		AEGIS_ERROR("Cannot open '%s' (%s)", storage_file_name.c_str(),
					   strerror(errno));
		return(errno);
	}
}


extern "C" {
int
aegis_certman_has_private_key_by_id(aegis_key_id key_id)
{
	string storage_file_name;
	struct stat fs;
	int rc;

	local_storage_dir(storage_file_name, priv_keys_dir);
	append_hex(storage_file_name, key_id, AEGIS_KEY_ID_LEN);
	storage_file_name.append(".pem");
	rc = stat(storage_file_name.c_str(), &fs);
	if (0 == rc) {
		AEGIS_DEBUG(1, "Has private key");
		return(1);
	} else {
		AEGIS_DEBUG(1, "Does not have a private key");
		return(0);
	}
}

int
aegis_certman_has_private_key(X509* cert)
{
	aegis_key_id key_id;

	if (NULL == cert || (0 != aegis_certman_get_key_id(cert, key_id)))
		return(0);
	else
		return(aegis_certman_has_private_key_by_id(key_id));
}
} // extern "C"

static int
read_key_from_file(aegis_key_id key_id, EVP_PKEY** key, char* passwd)
{
	string storage_file_name;
	BIO* infile;
	X509_SIG *p8 = NULL;
	PKCS8_PRIV_KEY_INFO *p8inf = NULL;
	int rc = 0;
	const char* lpasswd;

	if (passwd)
		lpasswd = passwd;
	else
		lpasswd = "";

	local_storage_dir(storage_file_name, priv_keys_dir);
	create_directory(storage_file_name.c_str(), PRIVATE_DIR_MODE);
	append_hex(storage_file_name, key_id, AEGIS_KEY_ID_LEN);
	storage_file_name.append(".pem");

	AEGIS_DEBUG(1, "Reading file '%s'", storage_file_name.c_str());

	infile = BIO_new_file(storage_file_name.c_str(), "rb");
	if (infile) {
		p8 = PEM_read_bio_PKCS8(infile, NULL, NULL, NULL);
		AEGIS_DEBUG(1, "PEM_read_bio_PKCS8 ret %p", p8);
		if (p8) {
			AEGIS_DEBUG(1, "Decrypting with '%s'", lpasswd);
			p8inf = (PKCS8_PRIV_KEY_INFO*)
				PKCS12_item_decrypt_d2i(p8->algor, 
										ASN1_ITEM_rptr(PKCS8_PRIV_KEY_INFO),
										lpasswd, 
										strlen(lpasswd),
										p8->digest,
										1);
			AEGIS_DEBUG(1, "PKCS8_decrypt ret %p", p8inf);
			X509_SIG_free(p8);
			if (p8inf) {
				*key = EVP_PKCS82PKEY(p8inf);
				if (*key) {
					AEGIS_DEBUG(1, "Retrieved key from '%s'", 
								   storage_file_name.c_str());
					rc = 0;
				} else {
					AEGIS_ERROR("Cannot retrieve key from '%s'", 
								   storage_file_name.c_str());
					rc = EINVAL;
				}
				PKCS8_PRIV_KEY_INFO_free(p8inf);
			} else 
				rc = EACCES;
		}
		else {
			AEGIS_ERROR("Cannot retrieve key from '%s'", storage_file_name.c_str());
			rc = EINVAL;
		}
		BIO_free(infile);
		return(rc);
	} else {
		AEGIS_ERROR("Cannot open '%s' (%s)", 
					   storage_file_name.c_str(),
					   strerror(errno));
		return(errno);
	}
}

#if 0
static void
remove_key_file(aegis_key_id key_id)
{
	string storage_file_name;
	int rc = 0;

	local_storage_dir(storage_file_name, priv_keys_dir);
	create_directory(storage_file_name.c_str(), PRIVATE_DIR_MODE);
	append_hex(storage_file_name, key_id, AEGIS_KEY_ID_LEN);
	storage_file_name.append(".pem");

	if (file_exists(storage_file_name.c_str())) {
		rc = unlink(storage_file_name.c_str());
		AEGIS_DEBUG(1, "Removed private key file '%s'", 
					   storage_file_name.c_str());
	}
}
#endif


static void
aegis_certman_int_init(void)
{
	string my_app_name;

    AEGIS_ENTER;
    if (!is_initialized) {
        process_name(my_app_name);
        AEGIS_DEBUG(1, "Init '%s'", my_app_name.c_str());
        aegis_crypto_init();
        is_initialized = 1;
    }
}


static int
local_iterate_storage_names(storage::visibility_t of_visibility, 
							storage::protection_t of_protection, 
							const char* matching_names,
							aegis_callback* cb_func,
							void* ctx)
{
	return(storage::iterate_storage_names(of_visibility, 
								 of_protection, 
								 matching_names,
								 cb_func, 
								 ctx));
}


// Visible part
extern "C" {

	struct cb_relay_par {
		void* o_ctx;
		aegis_callback* cb_func;
	};


	int 
	aegis_certman_open(X509_STORE** my_cert_store)
	{
		AEGIS_DEBUG(1, "%s: enter", __func__);
		aegis_certman_int_init();
		if (my_cert_store)
			*my_cert_store = X509_STORE_new();
		AEGIS_DEBUG(1, "%s: exit", __func__);
		return(0);
	}

	/*
	 * Some libraries call these functions without initializing
	 * the library first. Use this ugly hack to help them work
	 * at the moment.
	 */

#define AUTOINIT do {                                   \
		if (!is_initialized)							\
			aegis_certman_open(NULL);                   \
	} while (0)
	

	static int
	dont_change_private_stores_as_root(struct local_domain* in_domain)
	{
		if (0 == geteuid() 
			&& AEGIS_CERTMAN_DOMAIN_PRIVATE == in_domain->domflags) 
		{
			AEGIS_ERROR("changing private store '%s' as root", 
						   in_domain->dirname.c_str());
			return(1);
		} else
			return(0);
	}


	int aegis_certman_collect(const char* domain, int shared, X509_STORE* my_cert_store)
	{
		vector<string> x;
		const char* sep, *start = domain;
		storage::visibility_t storvis;
		int rc = 0;

		AUTOINIT;

		do {
			string domainname, dirname, storagename;
				
			sep = strchr(start, ':');
			if (sep) {
				domainname.assign(start, sep - start);
				start = sep + 1;
			} else
				domainname.assign(start);

			if (shared) {
				storvis = storage::vis_global;
				decide_storage_name(domainname.c_str(), AEGIS_CERTMAN_DOMAIN_SHARED, 
									dirname, storagename);
			} else {
				storvis = storage::vis_shared;
				decide_storage_name(domainname.c_str(), AEGIS_CERTMAN_DOMAIN_PRIVATE, 
									dirname, storagename);
			}

			if (directory_exists(dirname.c_str())) {
				storage* store = new storage(storagename.c_str(),
                                             cur_creds,
											 storvis, 
											 storage::prot_signed);
				storage::stringlist certs;
				int pos;
				
				AEGIS_DEBUG(1, "New store %p", store);
				pos = store->get_files(certs);
				AEGIS_DEBUG(1, "Check %d certificates", pos);
				for (int i = 0; i < pos; i++) {
					/* Never mind link files
					 */
					if (!store->contains_link(certs[i])) {
						if (store->verify_file(certs[i])) {
							AEGIS_DEBUG(1, "Load '%s'", certs[i]);
							x.push_back(certs[i]);
						} else
							AEGIS_ERROR("'%s' fails verification", certs[i]);
					}
				}
                store->release(certs);
				delete(store);
			} else {
				AEGIS_ERROR("'%s' does not exist", storagename.c_str());
				rc = EINVAL;
			}
		} while (sep);

		if (x.size()) {
			load_certs(x, true, my_cert_store);
		}

		return rc;
	}


	int
	aegis_certman_close(X509_STORE* my_cert_store)
	{
		AUTOINIT;

		AEGIS_DEBUG(1, "%s: enter", __func__);
		if (my_cert_store) {
			AEGIS_DEBUG(1, "%s: release %p", __func__, my_cert_store);
			X509_STORE_free(my_cert_store);
		}
		// bb5_finish();
		AEGIS_DEBUG(1, "%s: exit", __func__);
        if (cur_creds) {
            free(cur_creds);
            cur_creds = NULL;
        }
		return(0);
	}


	int 
	aegis_certman_open_domain(const char* domain_name, 
                                 int flags, 
                                 domain_handle* handle)
	{
		string storename;
		struct local_domain mydomain;
		storage::visibility_t storvis;
		int rc;

		AUTOINIT;

		*handle = NULL;
		mydomain.domflags = flags;
		decide_storage_name(domain_name, flags, mydomain.dirname, storename);

		if (AEGIS_CERTMAN_DOMAIN_PRIVATE == flags) {
			rc = create_directory(mydomain.dirname.c_str(), PRIVATE_DIR_MODE);
			storvis = storage::vis_shared;
		} else {
			rc = create_directory(mydomain.dirname.c_str(), PUBLIC_DIR_MODE);
			storvis = storage::vis_global;
		}

		if (0 != rc) {
			return(rc);
		}
		mydomain.index = new storage(storename.c_str(), 
                                     cur_creds, 
                                     storvis, 
									 storage::prot_signed);
		if (mydomain.index) {
			*handle = new struct local_domain(mydomain);
			return(0);
		} else
			return(-1);
	}


	int 
	aegis_certman_iterate_certs(
		domain_handle the_domain, 
		int cb_func(int,X509*,void*), 
		void* ctx)
	{
		storage::stringlist files;
		struct local_domain* mydomain;
		int i, pos, res = 0;

		AUTOINIT;

		if (!the_domain || !cb_func)
			return(-EINVAL);

		mydomain = (struct local_domain*)the_domain;
		pos = mydomain->index->get_files(files);
		AEGIS_DEBUG(1, "%s: domain contains %d objects", __func__, pos);
		for (i = 0; i < pos; i++) {
            /*
             * Skip link files
             */
            if (mydomain->index->contains_link(files[i])) {
                AEGIS_DEBUG(1, "%s: %s is a link, never mind", __func__, 
                            files[i]);
                continue;
            }

			X509* cert = load_cert_from_file(files[i]);
			if (cert) {
				res = cb_func(i, cert, ctx);
				AEGIS_DEBUG(5, "callback returned %d", res);
				if (res != -1)
					X509_free(cert);
				else
					res = 0;
				if (res)
					break;
			}
		}
		mydomain->index->release(files);	/* to prevent the memory leaks */
		return(res);
	}


	int aegis_certman_load_cert(domain_handle the_domain, 
								   aegis_key_id with_id, 
								   X509** cert)
	{
		
		string filename;
		struct local_domain *my_domain = (struct local_domain*)the_domain;

		AUTOINIT;

		filename = my_domain->dirname;
		filename.append(PATH_SEP);
		append_hex(filename, with_id, AEGIS_KEY_ID_LEN);
		filename.append(".pem");
		AEGIS_DEBUG(1, "Retrieve cert from '%s'", filename.c_str());
		// TODO: check integrity
		*cert = load_cert_from_file(filename.c_str());
		if (*cert)
			return(0);
		else
			return(ENOENT);
	}


	int
	aegis_certman_nbrof_certs(domain_handle in_domain)
	{
		AUTOINIT;

		if (in_domain) {
            storage *cert_store = ((struct local_domain*)in_domain)->index;
            if (cert_store)
                return cert_store->nbrof_files() - cert_store->nbrof_links();
            else
                return 0;
		} else
			/*
			 * No domain, no certs
			 */
			return(0);
	}

    /* A quick hack to make loading many certificates a tad faster
     */
    static int do_commit = 1;

	int 
	aegis_certman_add_cert(domain_handle to_domain, X509* cert)
	{
		struct local_domain* mydomain = (struct local_domain*)to_domain;
		FILE* to_file;
		string filename;
		int rc = 0;

		AUTOINIT;

		if (!to_domain || !cert)
			return EINVAL;
			
		if (V_ASN1_NEG_INTEGER == X509_get_serialNumber(cert)->type) {
			AEGIS_ERROR("Certificate has negative serial number");
			return EINVAL;
		}

		if (dont_change_private_stores_as_root(mydomain))
			return EACCES;

		if (!make_unique_filename(cert, mydomain, filename))
			return EEXIST;
		
		to_file = fopen(filename.c_str(), "w+");
		if (to_file) {
			if (PEM_write_X509(to_file, cert)) {
				AEGIS_DEBUG(1, "written %s", filename.c_str());
			} else {
				AEGIS_DEBUG(1, "cannot write to %s (%s)", filename.c_str(), 
					  strerror(errno));
				rc = errno;
			}
			fclose(to_file);
		} else
			rc = errno;

		if (0 == rc) {
            string hash_name;
			make_hash_filename(cert, mydomain->index, filename.c_str(), 
							   hash_name);
            if (!mydomain->index->contains_file(filename.c_str())) {
                mydomain->index->add_file(filename.c_str());
                if ("" != hash_name)
                    mydomain->index->add_link(hash_name.c_str(), 
											  filename.c_str());
                if (do_commit) {
					if (!mydomain->index->commit()) {
						AEGIS_DEBUG(1, "%s: add of '%s' failed (%s)", __func__, 
									filename.c_str(), strerror(errno));
						rc = errno;
                        unlink(filename.c_str());
					} else if (!mydomain->index->contains_file(filename.c_str())) {
						AEGIS_DEBUG(1, "%s: add of '%s' failed mysteriously", __func__, 
									filename.c_str());
                        unlink(filename.c_str());
                        rc = EACCES;
                    }
                }
            }
		}
		return rc;
	}

	int
	aegis_certman_add_certs(domain_handle to_domain, 
                            char* cert_files[], 
                            unsigned count)
	{
		vector<string> certs;
		X509_STORE* tmp_store;
		X509_OBJECT* obj;
		int rc, added = 0, l_errno = 0;

		AUTOINIT;

		if (!to_domain) {
			errno = EINVAL;
			return -1;
		}

		if (dont_change_private_stores_as_root((struct local_domain*)to_domain)) {
			errno = EACCES;
			return -1;
		}

		for (unsigned i = 0; i < count; i++) {
			if (cert_files[i]) {
				if (file_exists(cert_files[i])) {
					certs.push_back(cert_files[i]);
				} else
					AEGIS_ERROR("Invalid certificate file '%s'", cert_files[i]);
			} else
				break;
		}
		tmp_store = X509_STORE_new();
		if (tmp_store) {
			if (load_certs(certs, false, tmp_store)) {
                do_commit = 0;
				for (int i = 0; i < sk_X509_OBJECT_num(tmp_store->objs); i++) {
					obj = sk_X509_OBJECT_value(tmp_store->objs, i);
					if (X509_LU_X509 == obj->type) {
						rc = aegis_certman_add_cert(to_domain, obj->data.x509);
						if (0 == rc)
							added++;
						else {
							AEGIS_DEBUG(1, "%s: failed to add a certificate (%s)",
										__func__, strerror(rc));
							l_errno = rc;
						}
					}
				}
                do_commit = 1;
				/* Force commit in the build environment even if the store
				 * would not have changed.
				 */
                if (0 < added || aegis_system_plain == aegis_current_mode()) {
                    if (!((struct local_domain*)to_domain)->index->commit()) {
						AEGIS_ERROR("%s: add certs failed (%s)", __func__,
									strerror(errno));
						added = -1;
					}
				}
			} else {
				AEGIS_ERROR("Loading certificates failed");
			}
			X509_STORE_free(tmp_store);
		} else {
			AEGIS_ERROR("Cannot create X509_STORE");
		}
		if (0 != l_errno)
			errno = l_errno;
		return added;
	}


	int
	aegis_certman_rm_cert(domain_handle from_domain, aegis_key_id key_id)
	{
		string filename;
		struct local_domain* mydomain = (struct local_domain*)from_domain;
		int removed = 0;

		AUTOINIT;

		/*
		 * Allow removing certificates as root
		 */
		if (!mydomain) {
			return EINVAL;
		}

		filename = mydomain->dirname;
		filename.append(PATH_SEP);
		append_hex(filename, key_id, AEGIS_KEY_ID_LEN);
		for (int i = 0; i < 1000; i++) {
			string test_name(filename);
			if (0 < i) {
				char onbr[10];
				sprintf(onbr, "-%d", i);
				test_name.append(onbr);
			}
			test_name.append(".pem");
			if (mydomain->index->contains_file(test_name.c_str())) {
				AEGIS_DEBUG(1, "Remove cert file '%s'", test_name.c_str());
				mydomain->index->remove_file(test_name.c_str());
				if (0 == mydomain->index->contains_file(test_name.c_str())) {
					removed++;
					if (0 != unlink(test_name.c_str())) {
						if (0 == errno)
							errno = EACCES;
						break;
					}
					errno = 0;
				} else {
					AEGIS_DEBUG(1, "%s: couldn't remove '%s' (%s)", 
								__func__, test_name.c_str(), strerror(errno));
					return errno;
				}
			} else
				/* No such file */
				break;
		}

		if (0 == removed) {
			if (0 == errno)
				return ENOENT;
			else
				return errno;
		}

		/* Do not remove keys in case they are still used
		 * for another purpose.
		 */
		if (!mydomain->index->commit()) {
			AEGIS_DEBUG(1, "%s: rm cert failed (%s)", __func__,
						strerror(errno));
			return errno;
		}
		return 0;
	}


	int 
	aegis_certman_close_domain(domain_handle handle)
	{
		struct local_domain* mydomain;

		AUTOINIT;

		if (!handle)
			return EINVAL;
		mydomain = (struct local_domain*)handle;
		delete(mydomain->index);
		delete(mydomain);
		return(0);
	}


	int 
	aegis_certman_domain_changed(domain_handle handle)
	{
		struct local_domain* mydomain;

		AUTOINIT;

		if (!handle)
			return 0;
		mydomain = (struct local_domain*)handle;
		if (mydomain->index->refresh())
			return 1;
		else
			return 0;
	}


	int 
	aegis_certman_get_key_id(X509* of_cert, aegis_key_id to_this)
	{
		AUTOINIT;

		if (!of_cert && !to_this)
			return(EINVAL);
		if (X509_pubkey_digest(of_cert, EVP_sha1(), to_this, NULL))
			return(0);
		else
			return(EINVAL);
	}

	/*
	 * Nickname is constructed by catenating these name components
	 * together in the order of precedence. -1 is a group separator.
	 */
	static const int nickname_components [] = {
		NID_commonName, -1, 
		NID_organizationalUnitName, NID_organizationName, NID_countryName, -1,
		NID_organizationalUnitName, NID_organizationName, -1,
		NID_organizationName, NID_countryName, -1,
		NID_organizationName, -1, 
		NID_organizationalUnitName, -1,
		-1
	};

	int
	aegis_certman_get_nickname(X509* of_cert, 
								  char* to_buf, 
								  unsigned buf_len)
	{
		X509_NAME* name;
		int idx, npos = 0;
		string result;

		if (NULL == of_cert || NULL == to_buf || 0 == buf_len)
			return(EINVAL);

		name = X509_get_subject_name(of_cert);
		if (NULL == name)
			return(EINVAL);

		while (-1 != nickname_components[npos]) {
			idx = X509_NAME_get_index_by_NID(name, nickname_components[npos], -1);
			if (-1 == idx) {
				/*
				 * Component not found, roll forward
				 */
				result = "";
				while (-1 != nickname_components[npos])
					npos++;
			} else {
				X509_NAME_ENTRY *entry;
				ASN1_STRING *str = NULL;
				unsigned char *quark = NULL;

				entry = X509_NAME_get_entry(name, idx);
				if (NULL != entry)
					str = X509_NAME_ENTRY_get_data(entry);
				if (NULL != str)
					ASN1_STRING_to_UTF8(&quark, str);
				if (NULL != quark) {
					if ("" != result)
						result.append("/");
					result.append((char*)quark);
					OPENSSL_free(quark);
				}
			}
			npos++;
		}

		if (result.length() < buf_len)
			strcpy(to_buf, result.c_str());
		else {
			memcpy(to_buf, result.c_str(), buf_len);
			to_buf[buf_len] = '\0';
		}
		return(0);
	}


	int 
	aegis_certman_store_key(aegis_key_id with_id, 
							   EVP_PKEY* the_key, 
							   char* with_passwd)
	{
		AUTOINIT;
		if (0 == getuid()) {
			AEGIS_ERROR("adding private key as root");
		}
		return(store_key_to_file(with_id, the_key, with_passwd));
	}


	int
	aegis_certman_retrieve_key(aegis_key_id with_id, 
							   EVP_PKEY** the_key, 
							   char* with_passwd)
	{
		AUTOINIT;

		return(read_key_from_file(with_id, the_key, with_passwd));
	}


	static int
	cb_relay_key(int ord_nr, void* filename, void* ctx)
	{
		aegis_key_id key_id;
		struct cb_relay_par *pars = (struct cb_relay_par*) ctx;
		hex_to_key_id((const char*)filename, key_id);
		return(pars->cb_func(ord_nr, key_id, pars->o_ctx));
	}


	int
	aegis_certman_iterate_keys(aegis_callback* cb_func, void* ctx)
	{
		string keystore_name;
		string name_expression;
		struct cb_relay_par relay_pars;

		AUTOINIT;

		name_expression = "^";
		for (int i = 0; i < AEGIS_KEY_ID_LEN; i++)
			name_expression.append("[0-9a-f][0-9a-f]");
		name_expression.append("\\.pem$");
		local_storage_dir(keystore_name, priv_keys_dir);
		relay_pars.o_ctx = ctx;
		relay_pars.cb_func = cb_func;
		return(iterate_files(keystore_name.c_str(), 
							 name_expression.c_str(), 
							 cb_relay_key, 
							 &relay_pars));
	}


	static int
	cb_relay_storename(int ord_nr, void* filename, void* ctx)
	{
		struct cb_relay_par *pars = (struct cb_relay_par*) ctx;
		char* logical_name = (char*) filename;
		
		if (strlen(logical_name) > strlen(cert_storage_prefix))
			logical_name += strlen(cert_storage_prefix) + 1;
		return(pars->cb_func(ord_nr, logical_name, pars->o_ctx));
	}


	int
	aegis_certman_iterate_domains(int flags,
									 aegis_callback* cb_func,
									 void* ctx)
	{
		storage::visibility_t vis;
		string storage_names = cert_storage_prefix;
		struct cb_relay_par relay_pars;

		AUTOINIT;

		storage_names.assign(cert_storage_prefix);
		storage_names.append("\\..*");
		if (AEGIS_CERTMAN_DOMAIN_SHARED == flags)
			vis = storage::vis_global;
		else if (AEGIS_CERTMAN_DOMAIN_PRIVATE == flags)
			vis = storage::vis_shared;
		else
			return(0 - EINVAL);
		relay_pars.o_ctx = ctx;
		relay_pars.cb_func = cb_func;
		AEGIS_DEBUG(1, "Iterating storages '%s'", storage_names.c_str());
		return(local_iterate_storage_names(vis, 
										   storage::prot_signed, 
										   storage_names.c_str(), 
										   cb_relay_storename, 
										   &relay_pars));
	}

	int 
	aegis_certman_key_id_to_str(aegis_key_id key_id, 
								   char* to_buf, 
								   unsigned max_len)
	{
		unsigned i;
		char* start = to_buf;
		
		AUTOINIT;

		if (max_len < AEGIS_KEY_ID_STR_LEN)
			return(EINVAL);

		for (i = 0; i < AEGIS_KEY_ID_LEN; i++) {
			sprintf(to_buf, "%02hx", key_id[i]);
			to_buf += strlen(to_buf);
		}
		AEGIS_DEBUG(1, "%s: %s", __func__, start);
		return(0);
	}


	int 
	aegis_certman_str_to_key_id(char* from_str, 
								   aegis_key_id key_id)
	{
		unsigned i = 0;
		unsigned short b;
		const char* f = from_str;

		AUTOINIT;

		if (!f)
			return(0);

		while (*f && sscanf(f, "%02hx", &b)) {
			f += 2;
			if (*f == ':')
				f++;
			key_id[i++] = (unsigned char)b;
			if (i == AEGIS_KEY_ID_LEN)
				break;
		}
		
		if (i < AEGIS_KEY_ID_LEN || *f) {
			return(EINVAL);
		} else
			return(0);
	}

	/*
	 * Debug
	 */
	int
	inspect_certificate(const char* pathname)
	{
		x509_container xc(pathname);
		return(0);
	}

    int 
    aegis_certman_set_credentials(const char* token_name)
    {
        if (NULL != cur_creds)
            free(cur_creds);
        cur_creds = strdup(token_name);
        return 0;
    }


} // extern "C"
