/* -*- mode:c++; tab-width:4; c-basic-offset:4; -*-
 *
 * This file is part of aegis-certman
 *
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
 *
 * Contact: Juhani Mäkelä <ext-juhani.3.makela@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

/// \cond Don't make doxygen documentation

#include "x509_container.h"
#include <aegis_certman.h>

namespace aegis {

	x509_container::x509_container(X509* cert)
	{
		m_handled = false;
		m_verified = false;
		m_cert = NULL;
		m_bio = NULL;

        m_cert = X509_dup(cert);
        analyze_cert();
    }

	x509_container::x509_container(const char* pathname)
	{
		FILE* fp = fopen(pathname, "r");

		m_handled = false;
		m_verified = false;
		m_cert = NULL;
		m_bio = NULL;

		AEGIS_DEBUG(2, "Loading %s", pathname);

		if (fp) {
			m_cert = PEM_read_X509_AUX(fp, NULL, 0, NULL);
			fclose(fp);
		
			if (m_cert) {
				if (V_ASN1_NEG_INTEGER == X509_get_serialNumber(m_cert)->type) {
					AEGIS_ERROR("Certificate has negative serial number");
					return;
				}
				analyze_cert();
			} else {
				AEGIS_ERROR("cannot load certificate from '%s'", pathname);
			}
		} else
			AEGIS_ERROR("cannot find file '%s' (%d)", pathname, errno);
	}


	x509_container::~x509_container()
	{
		if (m_cert)
			X509_free(m_cert);
		if (m_bio)
			BIO_free(m_bio);
	}


	bool
	x509_container::get_extension(int nid, string& to_buf)
	{
		char buf [255];
		X509_EXTENSION *ext;
		int pos, len;

		pos = X509_get_ext_by_NID(m_cert, nid, -1);
		if (-1 == pos) {
			AEGIS_DEBUG(2, "NID %d does not exist in this certificate", nid);
			return(false);
		}

		ext = sk_X509_EXTENSION_value(m_cert->cert_info->extensions, pos);
		if (NULL == ext) {
			AEGIS_ERROR("Attribute %d not found", pos);
			return(false);
		}

		if (NULL == m_bio) {
			m_bio = BIO_new(BIO_s_mem());
			if (!m_bio) {
				AEGIS_ERROR("cannot create new BIO");
				return(false);
			}
		}

		X509V3_EXT_print(m_bio, ext, 0, 0);
		len = BIO_gets(m_bio, buf, sizeof(buf));

		if (len && buf[len - 1] == '\n')
			buf[len - 1] = '\0';
		
		to_buf = buf;

		/*
		 * Flush the BIO if there is more data than fits
		 * the buffer.
		 */
		while (sizeof(buf) == BIO_gets(m_bio, buf, sizeof(buf)));

		return(true);
	}

	#define KEYID_PFIX "keyid:"

	void 
	x509_container::analyze_cert()
	{
		char name_buf[1024];
		bool self_signed;
		int error;

		m_subject_name = X509_NAME_oneline(X509_get_subject_name(m_cert), 
										   name_buf, sizeof(name_buf));

		AEGIS_DEBUG(1, "Analyzing '%s'", m_subject_name.c_str());

		m_issuer_name = X509_NAME_oneline(X509_get_issuer_name(m_cert), 
										  name_buf, sizeof(name_buf));

		if (!get_extension(NID_subject_key_identifier, m_key_id)) {

			AEGIS_DEBUG(1, "Subject key id not defined!");

			aegis_key_id key_id;
			if (0 == aegis_certman_get_key_id(m_cert, key_id)) {
				for (int i = 0; i < AEGIS_KEY_ID_LEN; i++) {
					char tmp[3];
					sprintf(tmp, "%02X", key_id[i]);
					if (i)
						m_key_id.append(":");
					m_key_id.append(tmp);
				}
			}
		}
		if (!get_extension(NID_authority_key_identifier, m_issuer_key_id)) {
			AEGIS_DEBUG(1, "Authority key id not defined");

		} else {
			/*
			 * Remove "keyid:" prefix if exists, otherwise set empty
			 */
			if (0 == memcmp(m_issuer_key_id.c_str(), KEYID_PFIX, strlen(KEYID_PFIX)))
				m_issuer_key_id.erase(0, strlen(KEYID_PFIX));
			else
				m_issuer_key_id = "";
		}
		self_signed = is_issued_by(m_cert, &error);
		if (self_signed) {
			m_issuer_key_id = m_key_id;
			AEGIS_DEBUG(2, "\nkey_id       = %s\nissuer_key_id= %s\nis self signed", 
						   m_key_id.c_str(), m_issuer_key_id.c_str());

		} else if (   X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT == error
				   || X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY == error)
			AEGIS_DEBUG(2, "\nkey_id       = %s\nissuer_key_id= %s\nnot self signed", 
						   m_key_id.c_str(), m_issuer_key_id.c_str());

		else {
			m_key_id = "";
			AEGIS_DEBUG(2, "\nkey_id       = %s\nissuer_key_id= %s\nInvalid (%d)", 
						   m_key_id.c_str(), m_issuer_key_id.c_str(), error);
		}
		BIO_free(m_bio);
		m_bio = NULL;
	}


	bool 
	x509_container::is_issued_by(X509* cert, int* error)
	{
		X509_STORE* tmp_store;
		X509_STORE_CTX *csc;
		bool retval;
		int rc;

		tmp_store = X509_STORE_new();
		X509_STORE_add_cert(tmp_store, cert);

		csc = X509_STORE_CTX_new();
		rc = X509_STORE_CTX_init(csc, tmp_store, m_cert, NULL);

		retval = (X509_verify_cert(csc) > 0);
		*error = csc->error;
		if (!retval) {
			AEGIS_DEBUG(1, "%s: verification fails: %s", __func__, 
						X509_verify_cert_error_string(csc->error));
		}
		X509_STORE_CTX_free(csc);
		X509_STORE_free(tmp_store);

		return(retval);
	}


	bool 
	x509_container::is_self_signed()
	{
		bool retval;
		int error;

		AEGIS_DEBUG(1, "Check if '%s' is self signed", 
					   m_subject_name.c_str());

		retval = is_issued_by(m_cert, &error);

		AEGIS_DEBUG(1, "'%s' %s self signed", 
					   m_subject_name.c_str(), 
					   retval?"is":"is not");

		if (retval && "" == m_issuer_key_id)
			m_issuer_key_id = m_key_id;

		return(retval);
	}


	void
	x509_container::set_issuer(x509_container* to_this)
	{
		m_issuer_name = to_this->m_subject_name;
		m_issuer_key_id = to_this->m_key_id;
	}


	void
	x509_container::print(void)
	{
		PEM_write_X509(stdout, m_cert);
	}

    static string*
    name_component(X509_NAME* of_name, const int nid)
    {
        string *result = NULL;
		int idx;

		idx = X509_NAME_get_index_by_NID(of_name, nid, -1);
		if (-1 != idx) {
			X509_NAME_ENTRY *entry = NULL;
			entry = X509_NAME_get_entry(of_name, idx);
			if (NULL != entry) {
				ASN1_STRING *str = NULL;
				str = X509_NAME_ENTRY_get_data(entry);
				if (NULL != str) {
					unsigned char *quark = NULL;
					ASN1_STRING_to_UTF8(&quark, str);
					if (NULL != quark) {
                        result = new string((const char*)quark);
						OPENSSL_free(quark);
					}
				}
			}
		}
        if (NULL != result)
            return result;
        else
            return new string("");
	}

    string*
    x509_container::subject_name_nid(int nid)
    {
        return name_component(X509_get_subject_name(m_cert), nid);
    }

    string*
    x509_container::issuer_name_nid(int nid)
    {
        return name_component(X509_get_issuer_name(m_cert), nid);
    }

#define NOT_AVAIL_STRING "N/A"

    static string*
    ASN1_time_to_localtime_str(ASN1_TIME* atime)
    {
        struct tm r;
        int cnt;
        time_t res;
        char to_buf[256];
        unsigned buf_size = sizeof(to_buf);

        ASN1_GENERALIZEDTIME* agtime;
        
        if (!atime || !atime->data) {
            AEGIS_ERROR("Empty time");
            return(new string(NOT_AVAIL_STRING));
        }
		
        agtime = ASN1_TIME_to_generalizedtime(atime, NULL);
        
        if (!agtime) {
            AEGIS_ERROR("Invalid time '%s'", atime->data);
            return(new string(NOT_AVAIL_STRING));
        }
        memset(&r, '\0', sizeof(struct tm));
        cnt = sscanf((char*)agtime->data, "%04d%02d%02d%02d%02d%02d",
                     &r.tm_year, &r.tm_mon, &r.tm_mday, 
                     &r.tm_hour, &r.tm_min, &r.tm_sec);
        ASN1_TIME_free(agtime);
        if (cnt < 6) {
            AEGIS_ERROR("Invalid time '%s'", agtime->data);
            return(new string(NOT_AVAIL_STRING));
        }
        AEGIS_DEBUG(1, "%04d-%02d-%02d %02d:%02d:%02d", 
                       r.tm_year, r.tm_mon, r.tm_mday, 
                       r.tm_hour, r.tm_min, r.tm_sec);
        r.tm_mon--;
        r.tm_year -= 1900;
        /*
         * Unfortunately there is not an easy way to convert UTC time
         * to time_t, as mktime expects local time. Subtract the
         * timezone manually.
         */
        res = mktime(&r);
        if ((time_t)-1 != res) {
            struct tm ltm;
            res -= timezone;
            localtime_r(&res, &ltm);
            // Use the ISO 8601 timestamp format
            strftime(to_buf, buf_size, "%F %H:%M:%S", &ltm);
            return(new string(to_buf));
        }
        return(new string(NOT_AVAIL_STRING));
    }

    string*
    x509_container::not_before_timestamp()
    {
        return ASN1_time_to_localtime_str(X509_get_notBefore(m_cert));
    }

    string*
    x509_container::not_after_timestamp()
    {
        return ASN1_time_to_localtime_str(X509_get_notAfter(m_cert));
    }

    bool
    x509_container::is_valid_timewise()
    {
        time_t now = time(NULL);

        if (0 <= X509_cmp_time(X509_get_notBefore(m_cert), &now)) {
            AEGIS_DEBUG(1, "Certificate is not yet valid");
            return false; // X509_V_ERR_CERT_NOT_YET_VALID
        }
        if (0 >= X509_cmp_time(X509_get_notAfter(m_cert), &now)) {
            AEGIS_DEBUG(1, "Certificate has expired");
            return false; // X509_V_ERR_CERT_HAS_EXPIRED
        }
        return true;
    }

    bool
    x509_container::is_ca()
    {
        return 0 != X509_check_ca(m_cert);
    }

}
/// \endcond
